@@ -1,4 +1,19 @@
-/* (c) 2003 David Lichteblau.  License: GNU GPL. */
+/* Copyright (c) 2003,2004,2005 David Lichteblau
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ */
 #include <curses.h>
 #include <signal.h>
 #include <term.h>
@@ -7,7 +22,10 @@
 static int
 compare(int (*handler)(tentry *, tentry *, LDAPMod **, void *),
 	void *userdata,
-	GArray *offsets, char *cleanname, char *dataname)
+	GArray *offsets,
+	char *cleanname,
+	char *dataname,
+	long *error_position)
 {
 	FILE *clean, *data;
 	int rc;
@@ -15,7 +33,8 @@ compare(int (*handler)(tentry *, tentry *, LDAPMod **, void *),
 
 	if ( !(clean = fopen(cleanname, "r+"))) syserr();
 	if ( !(data = fopen(dataname, "r"))) syserr();
-	rc = compare_streams(handler, userdata, offsets, clean, data, &pos);
+	rc = compare_streams(handler, userdata, offsets, clean, data, &pos,
+			     error_position);
 	if (fclose(clean) == EOF) syserr();
 	if (fclose(data) == EOF) syserr();
 
@@ -85,7 +104,7 @@ moddn(LDAP *ld, char *old, char *new, int dor, LDAPControl **ctrls)
 	int rc;
 	char **newrdns = ldap_explode_dn(new, 0);
 	char **ptr = newrdns;
-	char *newrdn = *ptr++;
+	char *newrdn = *ptr++; /* non-null (checked in validate_rename) */
 	GString *newsup = g_string_sized_new(strlen(new));
 	if (*ptr) g_string_append(newsup, *ptr++);
 	for (; *ptr; ptr++) {
@@ -102,6 +121,7 @@ struct ldapmodify_context {
 	LDAP *ld;
 	LDAPControl **controls;
 	int verbose;
+	int noquestions;
 };
 
 static int
@@ -139,7 +159,14 @@ ldapmodify_handler(tentry *clean, tentry *modified, LDAPMod **mods,
 		}
 	} else if (clean) {
 		if (verbose) printf("(delete) %s\n", entry_dn(clean));
-		if (ldap_delete_ext_s(ld, entry_dn(clean), ctrls, 0)) {
+		switch (ldap_delete_ext_s(ld, entry_dn(clean), ctrls, 0)) {
+		case 0:
+			break;
+		case LDAP_NOT_ALLOWED_ON_NONLEAF:
+			if (!ctx->noquestions)
+				return -2;
+			/* else fall through */
+		default:
 			ldap_perror(ld, "ldap_delete");
 			return -1;
 		}
@@ -332,7 +359,7 @@ cleanup:
 
 static LDAP *
 do_connect(char *server, char *user, char *password,
-	   int referrals, int starttls, int deref)
+	   int referrals, int starttls, int tls, int deref)
 {
 	LDAP *ld = 0;
 	int rc = 0;
@@ -344,6 +371,8 @@ do_connect(char *server, char *user, char *password,
 		strcpy(url + 7, server);
 		server = url;
 	}
+	if (ldap_set_option(ld, LDAP_OPT_X_TLS_REQUIRE_CERT, (void *) &tls))
+		ldaperr(ld, "ldap_set_option(LDAP_OPT_X_TLS)");
 	if ( rc = ldap_initialize(&ld, server)) {
 		fprintf(stderr, "ldap_initialize: %s\n", ldap_err2string(rc));
 		exit(1);
@@ -357,7 +386,8 @@ do_connect(char *server, char *user, char *password,
 		ldap_unbind_s(ld);
 		return 0;
 	}
-	if (ldap_set_option(ld, LDAP_OPT_REFERRALS, (void *) referrals))
+	if (ldap_set_option(ld, LDAP_OPT_REFERRALS,
+                            referrals ? LDAP_OPT_ON : LDAP_OPT_OFF))
 		ldaperr(ld, "ldap_set_option(LDAP_OPT_REFERRALS)");
 	if (ldap_set_option(ld, LDAP_OPT_DEREF, (void *) &deref))
 		ldaperr(ld, "ldap_set_option(LDAP_OPT_DEREF)");
@@ -418,7 +448,7 @@ save_ldif(GArray *offsets, char *clean, char *data,
 	fputs(name->str, s);
 	fputs("\n", s);
 
-	compare(ldif_handler, s, offsets, clean, data);
+	compare(ldif_handler, s, offsets, clean, data, 0);
 	if (fclose(s) == EOF) syserr();
 
 	printf("Your changes have been saved to %s.\n", name->str);
@@ -432,7 +462,7 @@ view_ldif(char *dir, GArray *offsets, char *clean, char *data)
 	char *name = append(dir, "/ldif");
 	if ( !(s = fopen(name, "w"))) syserr();
 	fputs("version: 1\n", s);
-	compare(ldif_handler, s, offsets, clean, data);
+	compare(ldif_handler, s, offsets, clean, data, 0);
 	if (fclose(s) == EOF) syserr();
 	view(name);
 	free(name);
@@ -465,10 +495,11 @@ analyze_changes(GArray *offsets, char *clean, char *data)
 {
 	struct statistics st;
 	int rc;
+	long pos;
 
 retry:
 	memset(&st, 0, sizeof(st));
-	rc = compare(statistics_handler, &st, offsets, clean, data);
+	rc = compare(statistics_handler, &st, offsets, clean, data, &pos);
 
 	/* Success? */
 	if (rc == 0) {
@@ -491,7 +522,7 @@ retry:
 	for (;;) {
 		switch (choose("What now?", "eQ?", "(Type '?' for help.)")) {
 		case 'e':
-			edit(data);
+			edit(data, pos);
 			goto retry;
 		case 'Q':
 			exit(0);
@@ -507,14 +538,15 @@ retry:
 
 static void
 commit(LDAP *ld, GArray *offsets, char *clean, char *data, LDAPControl **ctrls,
-       int verbose)
+       int verbose, int noquestions)
 {
 	struct ldapmodify_context ctx;
 	ctx.ld = ld;
 	ctx.controls = ctrls;
 	ctx.verbose = verbose;
+	ctx.noquestions = noquestions;
 	
-	switch (compare(ldapmodify_handler, &ctx, offsets, clean, data)) {
+	switch (compare(ldapmodify_handler, &ctx, offsets, clean, data, 0)) {
 	case 0:
 		puts("Done.");
 		exit(0);
@@ -563,6 +595,22 @@ skip(char *dataname, GArray *offsets)
 	FILE *s = fopen(dataname, "r");
 	if (!s) syserr();
 	read_entry(s, 0, &key, &entry, 0);
+        if (!entry) {
+                /* Im Normalfall wollen wir einen Eintrag in data
+                 * ueberspringen.  Wenn aber in data nichts mehr steht,
+                 * sind wir ueber die eigentlichen Aenderungen schon
+                 * hinweg und es handelt sich um eine Loeschung.  In
+                 * diesem Fall muessen wir nur das Offset aus der
+                 * Tabelle entfernen. */
+                int n;
+                for (n = 0; n < offsets->len; n++)
+                        if (g_array_index(offsets, long, n) >= 0) {
+                                g_array_remove_index(offsets, n);
+				break;
+			}
+                if (fclose(s) == EOF) syserr();
+                return;
+        }
 	entry_free(entry);
 	if ( (pos = ftell(s)) == -1) syserr();
 	if (fclose(s) == EOF) syserr();
@@ -629,38 +677,255 @@ static void
 offline_diff(char *a, char *b)
 {
 	GArray *offsets = read_offsets(a);
-	compare(ldif_handler, stdout, offsets, a, b);
+	compare(ldif_handler, stdout, offsets, a, b, 0);
 	g_array_free(offsets, 1);
 }
 
+void
+write_config(LDAP *ld, FILE *f, cmdline *cmdline)
+{
+	char *user = cmdline->user;
+	char *server = cmdline->server;
+	int limit;
+
+	if (!f) f = stdout;
+	fputs("# ldap.conf(5)\n", f);
+	fputs("# edit this as needed and paste into ~/.ldaprc\n", f);
+
+	/* URI/HOST */
+	fputc('\n', f);
+	fputs("# server name\n", f);
+	fputs("# (for parameterless operation, make sure to include at"
+	      " least this line)\n",
+	      f);
+	if (!server) 
+		ldap_get_option(ld, LDAP_OPT_URI, &server);
+	if (!server) 
+		ldap_get_option(ld, LDAP_OPT_HOST_NAME, &server);
+	if (server)
+		if (strstr(server, "://"))
+			fprintf(f, "URI %s\n", server);
+		else
+			fprintf(f, "HOST %s\n", server);
+
+	/* BASE */
+	fputc('\n', f);
+	fputs("# default search base\n", f);
+	if (cmdline->basedns->len) {
+		GPtrArray *basedns = cmdline->basedns;
+		int i;
+		if (basedns->len > 1)
+			fputs("### multiple namingcontexts found (uncomment"
+			      " one of these lines):\n",
+			      f);
+		for (i = 0; i < basedns->len; i++) {
+			if (basedns->len > 1) fputc('#', f);
+			fprintf(f, "BASE %s\n", g_ptr_array_index(basedns, i));
+		}
+	} else {
+		if (!cmdline->discover)
+			fputs("### no search base specified, retry with"
+			      " --discover?\n",
+			      f);
+		fputs("#BASE <dn>\n", f);
+	}
+	
+	/* BINDDN */
+	fputc('\n', f);
+	fputs("# user to bind as\n", f);
+	if (user && user[0] == '(')
+		user = find_user(ld, user);
+	if (user)
+		fprintf(f, "BINDDN %s\n", user);
+	else
+		fputs("#BINDDN <dn>\n", f);
+	
+	/* search options */
+	fputc('\n', f);
+	fputs("# search parameters (uncomment as needed)\n", f);
+	switch (cmdline->deref) {
+	case LDAP_DEREF_NEVER:
+		fputs("#DEREF never\n", f);
+		break;
+	case LDAP_DEREF_SEARCHING:
+		fputs("#DEREF searcing\n", f);
+		break;
+	case LDAP_DEREF_FINDING:
+		fputs("#DEREF finding\n", f);
+		break;
+	case LDAP_DEREF_ALWAYS:
+		fputs("#DEREF always\n", f);
+		break;
+	}
+	ldap_get_option(ld, LDAP_OPT_SIZELIMIT, &limit);
+	fprintf(f, "#SIZELIMIT %d\n", limit);
+	ldap_get_option(ld, LDAP_OPT_TIMELIMIT, &limit);
+	fprintf(f, "#TIMELIMIT %d\n", limit);
+}
+
+static struct ldap_objectclass *
+get_objectclass(GPtrArray *classes, char *name)
+{
+	int i;
+	char **ptr;
+	for (i = 0; i < classes->len; i++) {
+		struct ldap_objectclass *cls = g_ptr_array_index(classes, i);
+		if (!strcmp(cls->oc_oid, name))
+			return cls;
+		for (ptr = cls->oc_names; ptr && *ptr; ptr++)
+			if (!strcasecmp(*ptr, name))
+				return cls;
+	}
+	fprintf(stderr, "Error: Object class not found: %s\n", name);
+	exit(1);
+}
+
+static struct ldap_attributetype *
+get_attributetype(GPtrArray *types, char *name)
+{
+	int i;
+	char **ptr;
+	for (i = 0; i < types->len; i++) {
+		struct ldap_attributetype *cls = g_ptr_array_index(types, i);
+		if (!strcmp(cls->at_oid, name))
+			return cls;
+		for (ptr = cls->at_names; ptr && *ptr; ptr++)
+			if (!strcasecmp(*ptr, name))
+				return cls;
+	}
+	fprintf(stderr, "Error: Attribute type not found: %s\n", name);
+	exit(1);
+}
+
+static char *
+class_name(struct ldap_objectclass *cls)
+{
+	char **names = cls->oc_names;
+	if (names && *names)
+		return *names;
+	return cls->oc_oid;
+}
+
+static char *
+type_name(struct ldap_attributetype *at)
+{
+	char **names = at->at_names;
+	if (names && *names)
+		return *names;
+	return at->at_oid;
+}
+
+static void
+add_template(LDAP *ld,  FILE *s, GPtrArray *wanted, char *base)
+{
+	int i, j;
+	GPtrArray *classes = g_ptr_array_new();
+	GPtrArray *types = g_ptr_array_new();
+	GPtrArray *must = g_ptr_array_new();
+	GPtrArray *may = g_ptr_array_new();
+	char **ptr;
+	struct ldap_objectclass *cls;
+	struct ldap_attributetype *at;
+	struct ldap_objectclass *structural = 0;
+	
+	get_schema(ld, classes, types);
+	fputc('\n', s);
+
+	/* normalize "wanted" to oids */
+	for (i = 0; i < wanted->len; i++) {
+		char *name = g_ptr_array_index(wanted, i);
+		cls = get_objectclass(classes, name);
+		g_ptr_array_index(wanted, i) = cls->oc_oid;
+		if (cls->oc_kind == LDAP_SCHEMA_ABSTRACT)
+			fprintf(s, "### NOTE: objectclass is abstract: %s\n",
+				name);
+	}
+	/* add all superclasses */
+	for (i = 0; i < wanted->len; i++) {
+		cls = get_objectclass(classes, g_ptr_array_index(wanted, i));
+		for (ptr = cls->oc_sup_oids; ptr && *ptr; ptr++)
+			adjoin_str(wanted, *ptr);
+		if (cls->oc_kind == LDAP_SCHEMA_STRUCTURAL)
+			if (structural)
+				fprintf(s,
+					"### WARNING: extra structural object class: %s\n",
+					class_name(cls));
+			else {
+				fprintf(s,
+					"# structural object class: %s\n",
+					class_name(cls));
+				structural = cls;
+			}
+		for (ptr = cls->oc_at_oids_must; ptr && *ptr; ptr++) {
+			at = get_attributetype(types, *ptr);
+			g_ptr_array_remove(may, at);
+			for (j = 0; j < must->len; j++)
+				if (at == g_ptr_array_index(must, j))
+					break;
+			if (j >= must->len) g_ptr_array_add(must, at);
+		}
+		for (ptr = cls->oc_at_oids_may; ptr && *ptr; ptr++) {
+			at = get_attributetype(types, *ptr);
+			for (j = 0; j < must->len; j++)
+				if (at == g_ptr_array_index(must, j))
+					break;
+			if (j >= must->len) g_ptr_array_add(may, at);
+		}
+	}
+	if (!structural)
+		fputs("### WARNING: no structural object class specified!\n",
+		      s);
+
+	fprintf(s, "add %s\n", base ? base : "<DN>");
+	for (i = 0; i < wanted->len; i++) {
+		cls = get_objectclass(classes, g_ptr_array_index(wanted, i));
+		fprintf(s, "objectClass: %s\n", class_name(cls));
+	}
+	for (i = 0; i < must->len; i++) {
+		at = g_ptr_array_index(must, i);
+		if (strcmp(at->at_oid, "2.5.4.0"))
+			fprintf(s, "%s: \n", type_name(at));
+	}
+	for (i = 0; i < may->len; i++) {
+		at = g_ptr_array_index(may, i);
+		if (strcmp(at->at_oid, "2.5.4.0"))
+			fprintf(s, "#%s: \n", type_name(at));
+	}
+}
+
 int
 main(int argc, const char **argv)
 {
 	LDAP *ld;
-	char *server = 0;
-	char *user = 0;
-	char *password = 0;
-	char *base = 0;
-	int scope = LDAP_SCOPE_SUBTREE;
-	char *filter = "(objectclass=*)";
-	char **attrs = 0;
+	cmdline cmdline;
+	GPtrArray *ctrls = g_ptr_array_new();
 	static char dir[] = "/tmp/ldapvi-XXXXXX";
 	char *clean;
 	char *data;
 	GArray *offsets;
 	int changed;
-	int progress = 1;
-	int referrals = 1;
-	int add = 0;
 	FILE *s;
-	int managedsait = 0;
-	char *sortkeys = 0;
-	GPtrArray *ctrls = g_ptr_array_new();
 	FILE *target_stream;
-	int deref = LDAP_DEREF_NEVER;
-	int starttls = 0;
-	int verbose = 0;
-	int noquestions = 0;
+
+	cmdline.server = 0;
+	cmdline.basedns = g_ptr_array_new();
+	cmdline.scope = LDAP_SCOPE_SUBTREE;
+	cmdline.filter = 0;
+	cmdline.attrs = 0;
+	cmdline.user = 0;
+	cmdline.password = 0;
+	cmdline.progress = 1;
+	cmdline.referrals = 1;
+	cmdline.add = 0;
+	cmdline.managedsait = 0;
+	cmdline.sortkeys = 0;
+	cmdline.starttls = 0;
+	cmdline.tls = LDAP_OPT_X_TLS_TRY;
+	cmdline.deref = LDAP_DEREF_NEVER;
+	cmdline.verbose = 0;
+	cmdline.noquestions = 0;
+	cmdline.discover = 0;
+	cmdline.config = 0;
 
 	if (argc >= 2 && !strcmp(argv[1], "--diff")) {
 		if (argc != 4) {
@@ -671,23 +936,39 @@ main(int argc, const char **argv)
 		exit(0);
 	}
 
-	parse_arguments(argc, argv, &server, &base, &scope, &filter, &attrs,
-			&user, &password, &progress, &referrals, &add,
-			ctrls, &managedsait, &sortkeys, &starttls, &deref,
-			&verbose, &noquestions);
+	parse_arguments(argc, argv, &cmdline, ctrls);
 	target_stream = fixup_streams();
 
-	ld = do_connect(server, user, password, referrals, starttls, deref);
+	ld = do_connect(cmdline.server,
+			cmdline.user,
+			cmdline.password,
+			cmdline.referrals,
+			cmdline.starttls,
+			cmdline.tls,
+			cmdline.deref);
 	if (!ld) exit(1);
 	setupterm(0, 1, 0);
 
-	if (sortkeys)
-		append_sort_control(ld, ctrls, sortkeys);
+	if (cmdline.sortkeys)
+		append_sort_control(ld, ctrls, cmdline.sortkeys);
 	g_ptr_array_add(ctrls, 0);
+
+	if (cmdline.discover) {
+		if (cmdline.basedns->len > 0)
+			yourfault("Conflicting options given:"
+				  " --base and --discover.");
+		discover_naming_contexts(ld, cmdline.basedns);
+	}
+
+	if (cmdline.config) {
+		write_config(ld, target_stream, &cmdline);
+		exit(0);
+	}
+
 	if (target_stream) {
-		if (add) yourfault("Cannot --add entries noninteractively.");
-		search(target_stream, ld, base, scope, filter, attrs,
-		       (void *) ctrls->pdata, progress, 1);
+		if (cmdline.add)
+			yourfault("Cannot --add entries noninteractively.");
+		search(target_stream, ld, &cmdline, (void *) ctrls->pdata, 1);
 		exit(0);
 	}
 
@@ -699,21 +980,29 @@ main(int argc, const char **argv)
 	clean = append(dir, "/clean");
 	data = append(dir, "/data");
 	
-	if ( !(s = fopen(clean, "w"))) syserr();
+	if ( !(s = fopen(data, "w"))) syserr();
 	fputs("# ldapvi(1)\n", s);
-	if (add)
+	if (cmdline.add) {
+		if (cmdline.add->len) {
+			char *base = 0;
+			if (cmdline.basedns->len > 0)
+				base = g_ptr_array_index(cmdline.basedns, 0);
+			add_template(ld, s, cmdline.add, base);
+		}
+		if (fclose(s) == EOF) syserr();
+		cp("/dev/null", clean, 0, 0);
 		offsets = g_array_new(0, 0, sizeof(long));
-	else
-		offsets = search(s, ld, base, scope, filter, attrs,
-				 (void *) ctrls->pdata, progress, noquestions);
-	if (fclose(s) == EOF) syserr();
-	cp(clean, data, 0, 0);
-	edit(data);
+	} else {
+		offsets = search(s, ld, &cmdline, (void *) ctrls->pdata, 0);
+		if (fclose(s) == EOF) syserr();
+		cp(data, clean, 0, 0);
+	}
+	edit(data, 0);
 
-	if (noquestions) {
+	if (cmdline.noquestions) {
 		if (!analyze_changes(offsets, clean, data)) return 0;
 		commit(ld, offsets, clean, data, (void *) ctrls->pdata,
-		       verbose);
+		       cmdline.verbose, 1);
 		return 1;
 	}
 
@@ -725,12 +1014,14 @@ main(int argc, const char **argv)
 		switch (choose("Action?", "yqQvebrs?", "(Type '?' for help.)")) {
 		case 'y':
 			commit(ld, offsets, clean, data, (void *) ctrls->pdata,
-			       verbose);
+			       cmdline.verbose, 0);
 			changed = 1;
 			break; /* reached only on user error */
 		case 'q':
 			if (save_ldif(offsets, clean, data,
-				      server, user, managedsait))
+				      cmdline.server,
+				      cmdline.user,
+				      cmdline.managedsait))
 				break;
 			return 0;
 		case 'Q':
@@ -739,19 +1030,24 @@ main(int argc, const char **argv)
 			view_ldif(dir, offsets, clean, data);
 			break;
 		case 'e':
-			edit(data);
+			edit(data, 0);
 			changed = 1;
 			break;
 		case 'b':
-			user = login(ld, 0, 0, 1);
+			cmdline.user = login(ld, 0, 0, 1);
 			changed = 1; /* print stats again */
 			break;
 		case 'r':
 			ldap_unbind_s(ld);
 			ld = do_connect(
-				server, user, password, referrals,
-				starttls, deref);
-			printf("Connected to %s.\n", server);
+				cmdline.server,
+				cmdline.user,
+				cmdline.password,
+				cmdline.referrals,
+				cmdline.starttls,
+				cmdline.tls,
+				cmdline.deref);
+			printf("Connected to %s.\n", cmdline.server);
 			changed = 1; /* print stats again */
 			break;
 		case 's':
